"""
Basic Contents like TravelProtection, or Links
namespace: spider_base

"""

__all__ = [
    "LinkContent", "TravelProtection"
]

import logging
from datetime import timedelta

from django.db import models
from django.utils.html import escape
from django.shortcuts import redirect
from django.utils.translation import gettext
from django.middleware.csrf import CsrfViewMiddleware
from django.http.response import HttpResponseBase
from django.utils import timezone
from django.urls import reverse
from django.contrib.auth.hashers import (
    check_password, make_password
)
from django.utils.translation import gettext_lazy as _

from ..contents import BaseContent, add_content
from ..constants import (
    TravelLoginType, VariantType, ActionUrl
)

logger = logging.getLogger(__name__)

# raw and export: use references
link_abilities = frozenset(["add", "update", "update_raw", "raw", "export"])


@add_content
class PersistenceFeature(BaseContent):
    appearances = [
        {
            "name": "Persistence",
            "ctype": VariantType.component_feature.value,
            "strength": 0
        },
    ]

    class Meta:
        abstract = True

    @classmethod
    def feature_urls(cls, name):
        return [
            ActionUrl(
                reverse("spider_base:token-renew"),
                "renew-token"
            ),
            ActionUrl(
                reverse("spider_base:token-delete-request"),
                "delete-token"
            )
        ]


@add_content
class LinkContent(BaseContent):
    appearances = [{
        "name": "Link",
        "ctype": VariantType.raw_update.value
    }]
    expose_name = False
    expose_description = False

    content = models.ForeignKey(
        "spider_base.AssignedContent", related_name="+",
        on_delete=models.CASCADE,
        limit_choices_to={
            "strength_link__lte": 10,
        }
    )

    push = models.BooleanField(
        blank=True, default=False,
        help_text=_("Improve ranking of this Link.")
    )

    def get_content_name(self):
        return self.content.content.get_content_name()

    def get_content_description(self):
        return self.content.content.get_content_description()

    def get_strength(self):
        return self.content.content.get_strength()

    def get_priority(self):
        priority = self.content.content.get_priority()
        # pin to top
        if self.push and priority < 1:
            return 1
        return priority

    def get_strength_link(self):
        # don't allow links linking on links
        return 11

    def get_info(self):
        ret = self.content.content.get_info()
        return "%ssource=%s\x1elink\x1e" % (
            ret, self.associated.pk
        )

    def get_references(self):
        if not self.content:
            return []
        return [self.content]

    def get_form(self, scope):
        from ..forms import LinkForm
        if scope in link_abilities:
            return LinkForm
        # maybe not required
        return self.content.content.get_form(scope)

    def get_form_kwargs(self, **kwargs):
        if kwargs["scope"] in link_abilities:
            ret = super().get_form_kwargs(**kwargs)
            ret["uc"] = kwargs["uc"]
            ret["request"] = kwargs["request"]
        else:
            # maybe not required anymore
            ret = self.content.content.get_form_kwargs(**kwargs)
        return ret

    def access_add(self, **kwargs):
        _ = gettext
        kwargs["legend"] = escape(_("Create Content Link"))
        return super().access_add(**kwargs)

    def access_update(self, **kwargs):
        _ = gettext
        kwargs["legend"] = escape(_("Update Content Link"))
        return super().access_update(**kwargs)

    def access_raw_update(self, **kwargs):
        return redirect(
            'spider_base:ucontent-access',
            token=self.content.token,
            access='update'
        )

    def access(self, context):
        # context is updated and used outside!!
        # so make sure that func gets only a copy (**)
        context.setdefault("extra_outer_forms", [])
        func = self.access_default
        if context["scope"] in link_abilities:
            func = getattr(self, "access_{}".format(context["scope"]))

            # check csrf tokens manually
            csrferror = CsrfViewMiddleware().process_view(
                context["request"], None, (), {}
            )
            if csrferror is not None:
                # csrferror is HttpResponse
                return csrferror
            ret = func(**context)
            if context["scope"] == "update":
                # update function should never return HttpResponse
                #  (except csrf error)
                assert(not isinstance(ret, HttpResponseBase))
            return ret
        else:
            context["source"] = self
            context["uc"] = self.content.usercomponent
            ret = self.access(context)
            context["uc"] = self.usercomponent
            return ret


login_choices = [
    (TravelLoginType.none.value, _("No Login protection")),

    (TravelLoginType.fake_login.value, _("Fake login")),
    # TODO: to prevent circumventing deletion_period, tie to modified
    (TravelLoginType.wipe.value, _("Wipe")),
    (TravelLoginType.wipe_user.value, _("Wipe User")),
]


_login_protection = _(
    "No Login Protection: normal, default<br/>"
    "Fake Login: fake login and index<br/>"
    "Wipe: Wipe protected content, "
    "except they are protected by a deletion period<br/>"
    "Wipe User: destroy user on login"
)


class TravelProtectionManager(models.Manager):
    def get_active(self, now=None, no_stop=False):
        if not now:
            now = timezone.now()
        q = models.Q(active=True, start__lte=now)
        if not no_stop:
            q &= (models.Q(stop__isnull=True) | models.Q(stop__gte=now))
        return self.get_queryset().filter(q)


def default_start():
    return timezone.now()+timedelta(hours=3)


def default_stop():
    return timezone.now()+timedelta(days=7)


@add_content
class TravelProtection(BaseContent):
    appearances = [
        {
            "name": "TravelProtection",
            "strength": 10,
            # "ctype": VariantType.unique.value
        }
    ]

    objects = TravelProtectionManager()

    active = models.BooleanField(default=False)
    is_fake = models.BooleanField(default=False)
    start = models.DateTimeField(default=default_start, null=False)
    # no stop for no termination
    stop = models.DateTimeField(default=default_stop, null=True)

    login_protection = models.CharField(
        max_length=10, choices=login_choices,
        default=TravelLoginType.none.value, help_text=_login_protection
    )
    # use defaults from user
    hashed_secret = models.CharField(
        null=True, max_length=128
    )

    disallow = models.ManyToManyField(
        "spider_base.UserComponent", related_name="travel_protected",
        blank=True
    )

    def get_abilities(self, context):
        return ("deactivate",)

    def check_password(self, raw_password):
        """
        Return a boolean of whether the raw_password was correct. Handles
        hashing formats behind the scenes.
        """
        def setter(raw_password):
            self.hashed_secret = make_password(raw_password)
            self.save(update_fields=["hashed_secret"])
        return check_password(
            raw_password, self.hashed_secret, setter
        )

    def get_strength_link(self):
        return 5

    def get_priority(self):
        # pin to top with higher priority
        return 2

    def get_form(self, scope):
        from ..forms import TravelProtectionForm
        return TravelProtectionForm

    def get_form_kwargs(self, **kwargs):
        ret = super().get_form_kwargs(**kwargs)
        ret["uc"] = kwargs["uc"]
        ret["request"] = kwargs["request"]
        return ret

    def access_add(self, **kwargs):
        _ = gettext
        kwargs["legend"] = escape(_("Create Travel Protection"))
        return super().access_add(**kwargs)

    def access_update(self, **kwargs):
        _ = gettext
        kwargs["legend"] = escape(_("Update Travel Protection"))
        return super().access_update(**kwargs)

    def access_view(self, **kwargs):
        return ""

    def access_deactivate(self, **kwargs):
        if self.hashed_secret:
            if not self.check_password(
                kwargs["request"].GET.get("travel", "")
            ):
                return "failure"
        self.active = False
        self.save()
        return "success"



from django.urls import path


from .views import CreateEntry, VerifyEntry, HashAlgoView


app_name = "spider_verifier"

urlpatterns = [
    path(
        'task/<slug:task_id>/',
        CreateEntry.as_view(),
        name='task'
    ),
    path(
        'hash/<slug:hash>/',
        VerifyEntry.as_view(),
        name='verify'
    ),
    path(
        'hash/',
        HashAlgoView.as_view(),
        name='hash_algo'
    ),
    # NOTE: this view is csrf_exempted
    path(
        '',
        CreateEntry.as_view(),
        name='create'
    ),

]



__all__ = ["WebConfigForm"]

from django import forms
from .models import WebConfig


class WebConfigForm(forms.ModelForm):
    creation_url = forms.URLField(disabled=True, required=False)

    class Meta:
        model = WebConfig
        fields = ['config']

    def __init__(self, *, scope=None, user=None, **kwargs):
        super().__init__(**kwargs)
        self.fields["creation_url"].initial = self.instance.creation_url




import logging
import posixpath

from django.utils.html import escape
from django.db import models
from django.urls import reverse
from django.conf import settings
from django.template.loader import render_to_string
from django.utils.translation import gettext_lazy as _
from django.utils.translation import gettext
from django.http import HttpResponseRedirect
from django.core.files.storage import default_storage

from jsonfield import JSONField

from ranged_response import RangedFileResponse

from spkcspider.apps.spider.contents import BaseContent, add_content
from spkcspider.apps.spider.conf import (
    image_extensions, media_extensions
)
from spkcspider.apps.spider.helpers import (
    create_b64_token, prepare_description
)

from .conf import (
    LICENSE_CHOICES
)

logger = logging.getLogger(__name__)


# Create your models here.


def get_file_path(instance, filename):
    ret = getattr(settings, "FILE_FILET_DIR", "file_filet")
    size = getattr(settings, "FILE_FILET_SALT_SIZE", 45)
    # try 100 times to find free filename
    # but should not take more than 1 try
    # IMPORTANT: strip . to prevent creation of htaccess files or similar
    for _i in range(0, 100):
        ret_path = default_storage.generate_filename(
            posixpath.join(
                ret, str(instance.associated.usercomponent.user.pk),
                create_b64_token(size), filename.lstrip(".")
            )
        )
        if not default_storage.exists(ret_path):
            break
    else:
        raise Exception("Unlikely event: no free filename")
    return ret_path


class ContentWithLicense(BaseContent):
    license_name = models.CharField(
        max_length=255, null=False, default="other"
    )
    license_url = models.URLField(max_length=400, blank=True, null=True)
    sources = JSONField(default=list, blank=True)
    license_name_translation_list = LICENSE_CHOICES

    class Meta(BaseContent.Meta):
        abstract = True

    @property
    def full_license_name(self):
        return self.license_name_translation_list.get(
            self.license_name, {}
        ).get("name", self.license_name)

    def get_size(self):
        s = super().get_size()
        s += len(str(self.sources))
        return s


@add_content
class FileFilet(ContentWithLicense):
    expose_name = True
    expose_description = True

    appearances = [{"name": "File"}]
    file = models.FileField(upload_to=get_file_path, null=False, blank=False)

    def get_template_name(self, scope):
        if scope in ["add", "update"]:
            return 'spider_filets/file_form.html'
        return 'spider_base/view_form.html'

    def get_size(self):
        return self.file.size + super().get_size()

    def get_content_name(self):
        # in case no name is set
        return posixpath.basename(self.file.name)

    def get_info(self):
        ret = super().get_info()
        return "%sname=%s\x1e" % (ret, self.associated.name)

    def get_form(self, scope):
        from .forms import FileForm
        return FileForm

    def get_abilities(self, context):
        return set(("download",))

    def get_form_kwargs(self, **kwargs):
        ret = super().get_form_kwargs(**kwargs)
        ret["request"] = kwargs["request"]
        ret["uc"] = self.associated.usercomponent
        return ret

    def render_form(self, **kwargs):
        kwargs["enctype"] = "multipart/form-data"
        return super().render_form(**kwargs)

    def access_view(self, **kwargs):
        kwargs["object"] = self
        kwargs["associated"] = self.associated
        kwargs["type"] = None
        split = self.associated.name.rsplit(".", 1)
        if len(split) == 2:
            extension = split[1].lower()
            if extension in image_extensions:
                kwargs["type"] = "image"
            elif extension in media_extensions:
                kwargs["type"] = "media"
        if getattr(settings, "FILE_DIRECT_DOWNLOAD", False):
            kwargs["download"] = self.file.url
        else:
            kwargs["download"] = "{}?{}".format(
                reverse(
                    'spider_base:ucontent-access',
                    kwargs={
                        "token": self.associated.token,
                        "access": 'download'
                    }
                ), kwargs["spider_GET"].urlencode()
            )
        return (
            render_to_string(
                "spider_filets/file.html", request=kwargs["request"],
                context=kwargs
            ),
            ""
        )

    def access_download(self, **kwargs):
        if getattr(settings, "FILE_DIRECT_DOWNLOAD", False):
            response = HttpResponseRedirect(
                self.file.url,
            )
        else:
            response = RangedFileResponse(
                kwargs["request"],
                self.file.file,
                content_type='application/octet-stream'
            )
        name = self.associated.name
        if "." not in name:  # use ending of saved file
            ext = self.file.name.rsplit(".", 1)
            if len(ext) > 1:
                name = "%s.%s" % (name, ext[1])
        # name is sanitized to not contain \n, and other ugly control chars
        response['Content-Disposition'] = \
            'attachment; filename="%s"' % posixpath.basename(name.replace(
                r'"', r'\"'
            ))

        response["Access-Control-Allow-Origin"] = "*"
        return response

    def access_add(self, **kwargs):
        _ = gettext
        kwargs["legend"] = escape(_("Upload File"))
        return super().access_add(**kwargs)

    def access_update(self, **kwargs):
        _ = gettext
        kwargs["legend"] = escape(_("Update File"))
        return super().access_update(**kwargs)

    def save(self, *args, **kw):
        if self.pk is not None:
            orig = FileFilet.objects.get(pk=self.pk)
            if orig.file != self.file:
                orig.file.delete(False)
        super().save(*args, **kw)


@add_content
class TextFilet(ContentWithLicense):
    expose_name = "force"
    expose_description = True

    appearances = [{"name": "Text"}]

    editable_from = models.ManyToManyField(
        "spider_base.UserComponent", related_name="+",
        help_text=_("Allow editing from selected components."),
        blank=True
    )
    push = models.BooleanField(
        blank=True, default=False,
        help_text=_("Improve ranking of this content.")
    )

    text = models.TextField(default="", blank=True)

    def get_priority(self):
        # push to top
        if self.push:
            return 1
        return 0

    def get_template_name(self, scope):
        # view update form
        if scope == "update_guest":
            return 'spider_base/edit_form.html'
        elif scope == "view":
            return 'spider_base/text.html'
        return super().get_template_name(scope)

    def get_content_description(self):
        # use javascript instead
        # currently dead code
        return " ".join(
            prepare_description(
                self.text, 51
            )[:50]
        )

    def get_info(self):
        ret = super().get_info()
        return "%sname=%s\x1e" % (
            ret, self.associated.name

        )

    def get_size(self):
        return len(self.text.encode("utf8")) + super().get_size()

    def get_form(self, scope):
        if scope in ("raw", "export", "list"):
            from .forms import RawTextForm as f
        else:
            from .forms import TextForm as f
        return f

    def get_form_kwargs(self, **kwargs):
        ret = super().get_form_kwargs(**kwargs)
        ret["request"] = kwargs["request"]
        ret["source"] = kwargs.get("source", self.associated.usercomponent)
        ret["scope"] = kwargs["scope"]
        return ret

    def get_abilities(self, context):
        _abilities = set()
        source = context.get("source", self.associated.usercomponent)
        if self.id and self.editable_from.filter(
            pk=source.pk
        ).exists():
            _abilities.add("update_guest")
        return _abilities

    def access_update_guest(self, **kwargs):
        kwargs["legend"] = \
            escape(_("Update \"%s\" (guest)") % self.__str__())
        kwargs["inner_form"] = False
        return self.access_update(**kwargs)

    def access_view(self, **kwargs):

        kwargs["object"] = self
        kwargs["content"] = self.associated
        return (
            render_to_string(
                "spider_filets/text.html", request=kwargs["request"],
                context=kwargs
            ),
            ""
        )




from django.conf import settings
from django.db import models
from django.utils.translation import gettext_lazy as _
from django.contrib.auth.models import AbstractUser
from django.contrib.auth.models import UserManager as CoreUserManager


class UserManager(CoreUserManager):
    def create_superuser(self, username, password, email=None, **extra_fields):
        # fix users with optional email
        super().create_superuser(username, email, password, **extra_fields)


# Create your models here.

def default_quota_spider_user_local():
    return getattr(settings, "SPIDER_USER_QUOTA_LOCAL", None)


def default_quota_spider_user_remote():
    return getattr(settings, "SPIDER_USER_QUOTA_REMOTE", None)


def default_quota_spider_user_components():
    return getattr(settings, "SPIDER_USER_QUOTA_USERCOMPONENTS", None)


class SpiderUser(AbstractUser):
    """ A reference User Implementation suitable for spkcspider """
    REQUIRED_FIELDS = []
    SAFE_FIELDS = ['email', 'password']
    if getattr(settings, 'ALLOW_USERNAME_CHANGE', False):
        SAFE_FIELDS.insert(0, 'username')

    # optional quota
    quota_local = models.PositiveIntegerField(
        null=True, blank=True, default=default_quota_spider_user_local,
        help_text=_("Quota in Bytes, null for no limit")
    )
    quota_remote = models.PositiveIntegerField(
        null=True, blank=True, default=default_quota_spider_user_remote,
        help_text=_("Quota in Bytes, null for no limit")
    )
    quota_usercomponents = models.PositiveIntegerField(
        null=True, blank=True, default=default_quota_spider_user_components,
        help_text=_("Quota in units, null for no limit")
    )
    # first_name,last_name are used in the UserForm so don't remove them
    # unless you want spending your time adapting the adminform

    objects = UserManager()

    class Meta(AbstractUser.Meta):
        swappable = 'AUTH_USER_MODEL'



__all__ = ("filter_components", "filter_contents")

from django.db.models import Q
from django.conf import settings


def filter_components(searchlist, filter_unlisted=True, use_contents=True):
    searchq = Q()
    searchq_exc = Q()
    notsearch = Q()

    counter = 0
    # against ddos
    max_counter = settings.SPIDER_MAX_SEARCH_PARAMETERS

    # list only unlisted if explicity requested or export is used
    # ComponentPublicIndex doesn't allow unlisted in any case
    # this is enforced by setting "is_special_user" to False
    if filter_unlisted:
        notsearch = ~Q(contents__info__contains="\x1eunlisted\x1e")

    for item in searchlist:
        if filter_unlisted and item == "_unlisted":
            continue
        if counter > max_counter:
            break
        counter += 1
        if len(item) == 0:
            continue
        use_strict = False
        if item.startswith("!!"):
            _item = item[1:]
        elif item.startswith("__"):
            _item = item[1:]
        elif item.startswith("!_"):
            _item = item[2:]
            use_strict = True
        elif item.startswith("!"):
            _item = item[1:]
        elif item.startswith("_"):
            _item = item[1:]
            use_strict = True
        else:
            _item = item
        qob = Q()
        if use_strict:
            if use_contents:
                qob |= Q(contents__info__contains="\x1e%s\x1e" % _item)
                # exclude unlisted from searchterms
                qob &= notsearch
        else:
            qob |= Q(description__icontains=_item)
        if _item == "index":
            qob |= Q(strength=10)
        elif use_strict:
            qob |= Q(
                name=_item,
                strength__lt=10
            )
        else:
            qob |= Q(
                name__icontains=_item,
                strength__lt=10
            )
        if item.startswith("!!"):
            searchq |= qob
        elif item.startswith("!"):
            searchq_exc |= qob
        else:
            searchq |= qob
    return (searchq & ~searchq_exc, counter)


def filter_contents(
    searchlist, idlist, filter_unlisted=True, use_components=False
):
    searchq = Q()
    searchq_exc = Q()

    counter = 0
    unlisted_active = False
    # against ddos
    max_counter = settings.SPIDER_MAX_SEARCH_PARAMETERS

    for item in searchlist:
        if filter_unlisted is True and item == "_unlisted":
            continue
        elif item == "_unlisted":
            unlisted_active = True
        if counter > max_counter:
            break
        counter += 1
        if len(item) == 0:
            continue
        use_strict = False
        negate = False
        if item.startswith("!!"):
            _item = item[1:]
        elif item.startswith("__"):
            _item = item[1:]
        elif item.startswith("!_"):
            _item = item[2:]
            use_strict = True
            negate = True
        elif item.startswith("!"):
            _item = item[1:]
            negate = True
        elif item.startswith("_"):
            _item = item[1:]
            use_strict = True
        else:
            _item = item
        if use_strict:
            qob = Q(name=_item)
            qob |= Q(info__contains="\x1e%s\x1e" % _item)
            # can exclude/include specific usercomponents names
            if use_components:
                qob |= Q(usercomponent__name=_item)
        else:
            qob = Q(name__icontains=_item)
            qob |= Q(description__icontains=_item)
            qob |= Q(info__icontains=_item)
            # but don't be too broad with unspecific negation
            #   only apply on contents
            if use_components and not negate:
                qob |= Q(usercomponent__name__icontains=_item)
                qob |= Q(usercomponent__description__icontains=_item)
        if negate:
            searchq_exc |= qob
        else:
            searchq |= qob

    if idlist:
        # idlist contains int and str entries
        try:
            ids = map(lambda x: int(x), idlist)
        except ValueError:
            # deny any access in case of an incorrect id
            ids = []

        searchq &= (
            Q(
                id__in=ids
            )
        )
    if not unlisted_active:
        if filter_unlisted is True:
            searchq_exc |= Q(info__contains="\x1eunlisted\x1e")
        else:
            searchq_exc |= Q(
                info__contains="\x1eunlisted\x1e",
                priority__lte=filter_unlisted
            )
    return (searchq & ~searchq_exc, counter)



from django.contrib import admin
from django.contrib.auth import admin as user_admin
from django.utils.translation import gettext_lazy as _

from .models import SpiderUser
# Register your models here.


@admin.register(SpiderUser)
class UserAdmin(user_admin.UserAdmin):
    # exclude = ["first_name", "last_name"]
    list_display = ('username', 'email', 'is_active', 'is_staff')
    list_filter = ('is_staff', 'is_superuser', 'is_active', 'groups')
    search_fields = ('username', 'email')
    fieldsets = (
        (None, {'fields': ('username', 'password')}),
        (_('Personal info'), {'fields': ('email',)}),
        (_('Permissions'), {'fields': ('is_active', 'is_staff',
                                       'is_superuser',
                                       'quota_local', 'quota_remote', 'groups',
                                       'user_permissions')}),
        (_('Important dates'), {'fields': ('last_login', 'date_joined')}),
    )

    def has_module_permission(self, request):
        return True

    def has_change_permission(self, request, obj=None):
        if obj:
            if not request.user.is_active:
                return False
            if request.user.is_superuser or request.user == obj:
                return True
            # only superuser can alter superusers
            if obj.is_superuser:
                return False
        return super().has_change_permission(request, obj)

    def has_delete_permission(self, request, obj=None):
        if not request.user.is_active:
            return False
        if obj and obj != request.user:
            if request.user.is_superuser:
                return True
            # only superuser can delete superusers
            if obj.is_superuser:
                return False
        return super().has_delete_permission(request, obj)

    def get_readonly_fields(self, request, obj=None):
        fields = []
        if not request.user.is_superuser:
            fields.append("is_superuser")
            fields.append("last_login")
            fields.append("date_joined")
            if not self.has_change_permission(request, obj):
                fields.append("groups")
                fields.append("permissions")
                fields.append("is_staff")
        return fields



__all__ = {
    "validate", "valid_wait_states", "verify_download_size",
    "async_validate", "verify", "async_verify"
}

import logging
import binascii
import tempfile
import os

from django.utils.translation import gettext as _
from django.core.files import File
from django.conf import settings
from django.core import exceptions

from rdflib import Graph, URIRef, Literal
from rdflib.namespace import XSD
import requests
import certifi

from spkcspider import celery_app

from spkcspider.apps.spider.constants import spkcgraph
from spkcspider.apps.spider.helpers import merge_get_url, get_settings_func

from .constants import BUFFER_SIZE
from .functions import get_hashob
from .models import VerifySourceObject, DataVerificationTag

hashable_predicates = set([spkcgraph["name"], spkcgraph["value"]])

valid_wait_states = {
    "RETRIEVING", "HASHING", "STARTED"
}


def hash_entry(triple):
    h = get_hashob()
    if triple[2].datatype == XSD.base64Binary:
        h.update(triple[2].datatype.encode("utf8"))
        h.update(triple[2].toPython())
    else:
        if triple[2].datatype:
            h.update(triple[2].datatype.encode("utf8"))
        else:
            h.update(XSD.string.encode("utf8"))
        h.update(triple[2].encode("utf8"))
    return h.finalize()


def yield_hashes(graph, hashable_nodes):
    for t in graph.triples((None, spkcgraph["value"], None)):
        if (
            t[0] in hashable_nodes and
            t[2].datatype != spkcgraph["hashableURI"]
        ):
            yield hash_entry(t)


def yield_hashable_urls(graph, hashable_nodes):
    for t in graph.triples(
        (None, spkcgraph["value"], spkcgraph["hashableURI"])
    ):
        if t[0] in hashable_nodes:
            yield t


def verify_download_size(length, current_size=0):
    if not length or not length.isdigit():
        return False
    length = int(length)
    if settings.VERIFIER_MAX_SIZE_ACCEPTED < length:
        return False
    return True


def validate(ob, hostpart, task=None):
    dvfile = None
    source = None
    if isinstance(ob, tuple):
        dvfile = open(ob[0], "r+b")
        current_size = ob[1]
    else:
        current_size = 0
        dvfile = tempfile.NamedTemporaryFile(delete=False)

        source = VerifySourceObject.objects.get(
            id=ob
        )
        url = source.get_url()

        try:
            resp = requests.get(
                url, stream=True, verify=certifi.where(),
                timeout=settings.VERIFIER_REQUESTS_TIMEOUT
            )
        except requests.exceptions.Timeout:
            raise exceptions.ValidationError(
                _('url timed out: %(url)s'),
                params={"url": url},
                code="timeout_url"
            )
        except requests.exceptions.ConnectionError:
            raise exceptions.ValidationError(
                _('invalid url: %(url)s'),
                params={"url": url},
                code="invalid_url"
            )
        if resp.status_code != 200:
            raise exceptions.ValidationError(
                _("Retrieval failed: %(reason)s"),
                params={"reason": resp.reason},
                code="error_code:{}".format(resp.status_code)
            )

        c_length = resp.headers.get("content-length", None)
        if not verify_download_size(c_length, current_size):
            raise exceptions.ValidationError(
                _("Content too big: %(size)s"),
                params={"size": c_length},
                code="invalid_size"
            )
        c_length = int(c_length)
        current_size += c_length
        # preallocate file
        dvfile.truncate(c_length)
        dvfile.seek(0, 0)

        for chunk in resp.iter_content(BUFFER_SIZE):
            dvfile.write(chunk)
        dvfile.seek(0, 0)
    g = Graph()
    g.namespace_manager.bind("spkc", spkcgraph, replace=True)
    try:
        g.parse(
            dvfile.name,
            format="turtle"
        )
    except Exception:
        if settings.DEBUG:
            dvfile.seek(0, 0)
            logging.exception(dvfile.read())
        logging.error("Parsing file failed")
        dvfile.close()
        os.unlink(dvfile.name)
        raise exceptions.ValidationError(
            _('Invalid graph fromat'),
            code="invalid_format"
        )

    tmp = list(g.triples((None, spkcgraph["scope"], None)))
    if len(tmp) != 1:
        dvfile.close()
        os.unlink(dvfile.name)
        raise exceptions.ValidationError(
            _('Invalid graph'),
            code="invalid_graph"
        )
    start = tmp[0][0]
    scope = tmp[0][2].toPython()
    tmp = list(g.triples((start, spkcgraph["pages.num_pages"], None)))
    if len(tmp) != 1:
        dvfile.close()
        os.unlink(dvfile.name)
        raise exceptions.ValidationError(
            _('Invalid graph'),
            code="invalid_graph"
        )
    pages = tmp[0][2].toPython()
    tmp = list(g.triples((
        start,
        spkcgraph["pages.current_page"],
        Literal(1, datatype=XSD.positiveInteger)
    )))
    if len(tmp) != 1:
        dvfile.close()
        os.unlink(dvfile.name)
        raise exceptions.ValidationError(
            _('Must be page 1'),
            code="invalid_page"
        )
    if task:
        task.update_state(
            state='RETRIEVING',
            meta={
                'page': 1,
                'num_pages': pages
            }
        )
    tmp = list(g.objects(start, spkcgraph["action:view"]))
    if len(tmp) != 1:
        dvfile.close()
        os.unlink(dvfile.name)
        raise exceptions.ValidationError(
            _('Invalid graph'),
            code="invalid_graph"
        )
    view_url = tmp[0].toPython()
    if isinstance(ob, tuple):
        split = view_url.split("?", 1)
        source = VerifySourceObject.objects.update_or_create(
            url=split[0], defaults={"get_params": split[1]}
        )
    mtype = set()
    if scope == "list":
        mtype.add("UserComponent")
    else:
        mtype.update(map(
            lambda x: x.toPython(), g.objects(start, spkcgraph["type"])
        ))

    data_type = get_settings_func(
        "VERIFIER_CLEAN_GRAPH",
        "spkcspider.apps.verifier.functions.clean_graph"
    )(mtype, g, start, source, hostpart)
    if not data_type:
        dvfile.close()
        os.unlink(dvfile.name)
        raise exceptions.ValidationError(
            _('Invalid graph type: %(type)s'),
            params={"type": data_type},
            code="invalid_type"
        )

    # retrieve further pages
    for page in range(2, pages+1):
        url = merge_get_url(
            source.get_url(), raw="embed", page=str(page)
        )
        # validation not neccessary here (base url is verified)
        try:
            resp = requests.get(
                url, stream=True, verify=certifi.where(),
                timeout=settings.VERIFIER_REQUESTS_TIMEOUT
            )
        except requests.exceptions.Timeout:
            dvfile.close()
            os.unlink(dvfile.name)
            raise exceptions.ValidationError(
                _('url timed out: %(url)s'),
                params={"url": url},
                code="timeout_url"
            )
        except requests.exceptions.ConnectionError:
            dvfile.close()
            os.unlink(dvfile.name)
            raise exceptions.ValidationError(
                _('Invalid url: %(url)s'),
                params={"url": url},
                code="innvalid_url"
            )
        if resp.status_code != 200:
            dvfile.close()
            os.unlink(dvfile.name)
            raise exceptions.ValidationError(
                _("Retrieval failed: %(reason)s"),
                params={"reason": resp.reason},
                code="error_code:{}".format(resp.status_code)
            )

        c_length = resp.headers.get("content-length", None)
        if not verify_download_size(c_length, current_size):
            dvfile.close()
            os.unlink(dvfile.name)
            raise exceptions.ValidationError(
                _("Content too big: %(size)s"),
                params={"size": c_length},
                code="invalid_size"
            )
        c_length = int(c_length)
        current_size += c_length
        # clear file
        dvfile.truncate(c_length)
        dvfile.seek(0, 0)

        for chunk in resp.iter_content(BUFFER_SIZE):
            dvfile.write(chunk)
        dvfile.seek(0, 0)
        try:
            g.parse(
                dvfile.name,
                format="turtle"
            )
        except Exception as exc:
            if settings.DEBUG:
                dvfile.seek(0, 0)
                logging.error(dvfile.read())
            logging.exception(exc)
            dvfile.close()
            os.unlink(dvfile.name)
            # pages could have changed, but still incorrect
            raise exceptions.ValidationError(
                _("%(page)s is not a \"%(format)s\" file"),
                params={"format": "turtle", "page": page},
                code="invalid_file"
            )

        if task:
            task.update_state(
                state='RETRIEVING',
                meta={
                    'page': page,
                    'num_pages': pages
                }
            )

    hashable_nodes = set(g.subjects(
        predicate=spkcgraph["hashable"], object=Literal(True)
    ))

    hashes = [
        i for i in yield_hashes(g, hashable_nodes)
    ]
    if task:
        task.update_state(
            state='RETRIEVING',
            meta={
                'hashable_urls_checked': 0
            }
        )
    for count, t in enumerate(yield_hashable_urls(g, hashable_nodes), start=1):
        if (URIRef(t[2].value), None, None) in g:
            continue
        url = merge_get_url(t[2].value, raw="embed")
        if not get_settings_func(
            "SPIDER_URL_VALIDATOR",
            "spkcspider.apps.spider.functions.validate_url_default"
        )(url):
            dvfile.close()
            os.unlink(dvfile.name)
            raise exceptions.ValidationError(
                _('Insecure url: %(url)s'),
                params={"url": url},
                code="insecure_url"
            )

        try:
            resp = requests.get(
                url, stream=True, verify=certifi.where(),
                timeout=settings.VERIFIER_REQUESTS_TIMEOUT
            )
        except requests.exceptions.Timeout:
            dvfile.close()
            os.unlink(dvfile.name)
            raise exceptions.ValidationError(
                _('url timed out: %(url)s'),
                params={"url": url},
                code="timeout_url"
            )
        except requests.exceptions.ConnectionError:
            raise exceptions.ValidationError(
                _('Invalid url: %(url)s'),
                params={"url": url},
                code="innvalid_url"
            )
        if resp.status_code != 200:
            raise exceptions.ValidationError(
                _("Retrieval failed: %(reason)s"),
                params={"reason": resp.reason},
                code="code_{}".format(resp.status_code)
            )

        h = get_hashob()
        h.update(XSD.base64Binary.encode("utf8"))
        for chunk in resp.iter_content(BUFFER_SIZE):
            h.update(chunk)
        # do not use add as it could be corrupted by user
        # (user can provide arbitary data)
        g.set((
            URIRef(t[2].value),
            spkcgraph["hash"],
            Literal(h.finalize().hex())
        ))
        if task:
            task.update_state(
                state='RETRIEVING',
                meta={
                    'hashable_urls_checked': count
                }
            )
    if task:
        task.update_state(
            state='HASHING',
        )

    # make sure triples are linked to start
    # (user can provide arbitary data)
    g.remove((start, spkcgraph["hashed"], None))
    for t in g.triples((None, spkcgraph["hash"], None)):
        g.add((
            start,
            spkcgraph["hashed"],
            t[0]
        ))
        hashes.append(binascii.unhexlify(t[2].value))

    for i in g.subjects(spkcgraph["type"], Literal("Content")):
        h = get_hashob()
        h.update(i.encode("utf8"))
        hashes.append(h.finalize())
    hashes.sort()

    h = get_hashob()
    for i in hashes:
        h.update(i)
    # do not use add as it could be corrupted by user
    # (user can provide arbitary data)
    digest = h.finalize().hex()
    g.set((
        start,
        spkcgraph["hash"],
        Literal(digest)
    ))

    dvfile.truncate(0)
    dvfile.seek(0, 0)
    # save in temporary file
    g.serialize(
        dvfile, format="turtle"
    )

    result, created = DataVerificationTag.objects.get_or_create(
        defaults={
            "dvfile": File(dvfile),
            "source": source,
            "data_type": data_type
        },
        hash=digest
    )
    dvfile.close()
    os.unlink(dvfile.name)
    update_fields = set()
    # and source, cannot remove source without replacement
    if not created and source and source != result.source:
        result.source = source
        update_fields.add("source")
    if data_type != result.data_type:
        result.data_type = data_type
        update_fields.add("data_type")
    result.save(update_fields=update_fields)
    if task:
        task.update_state(
            state='SUCCESS'
        )
    return result


@celery_app.task(bind=True, name='async validation')
def async_validate(self, ob, hostpart):
    ret = validate(ob, hostpart, self)
    return ret.get_absolute_url()


def verify(tagid, task=None):
    tag = DataVerificationTag.objects.get(id=tagid)
    get_settings_func(
        "VERIFIER_TAG_VERIFIER",
        "spkcspider.apps.verifier.functions.verify_tag_default"
    )(tag)
    if tag.verification_state == "verified":
        try:
            tag.callback()
        except exceptions.ValidationError:
            logging.exception("Error while calling back")


@celery_app.task(bind=True, name='async verification', ignore_results=True)
def async_verify(self, tagid):
    verify(tagid, self)



# Generated by Django 2.1.5 on 2019-01-09 15:05

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0004_auto_20190103_0156'),
    ]

    operations = [
        migrations.RenameField(
            model_name='userinfo',
            old_name='used_space',
            new_name='used_space_local',
        ),
        migrations.AddField(
            model_name='userinfo',
            name='used_space_remote',
            field=models.BigIntegerField(default=0, editable=False),
        ),
    ]



# Generated by Django 2.1.5 on 2019-01-15 21:49

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0009_authtoken_persist'),
    ]

    operations = [
        migrations.AddField(
            model_name='usercomponent',
            name='can_auth',
            field=models.BooleanField(default=False, editable=False),
        ),
    ]



# Generated by Django 2.1.2 on 2018-10-31 15:06

from django.db import migrations, models
import spkcspider.apps.verifier.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='DataVerificationTag',
            fields=[
                ('id', models.BigAutoField(editable=False, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('hash', models.SlugField(max_length=512, unique=True)),
                ('dvfile', models.FileField(blank=True, help_text='File with data to verify', null=True, upload_to=spkcspider.apps.verifier.models.dv_path)),
                ('data_type', models.CharField(default='layout', max_length=20)),
                ('checked', models.DateTimeField(blank=True, null=True)),
                ('verification_state', models.CharField(choices=[('pending', 'pending'), ('verified', 'verified'), ('rejected', 'rejected')], default='pending', max_length=10)),
                ('note', models.TextField(default='')),
            ],
        ),
    ]



# Generated by Django 2.1.4 on 2019-01-02 22:57

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0002_20181231_squashed'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='usercomponent',
            name='avail_features',
        ),
    ]




__all__ = (
    "ProtectionType", "VariantType", "ProtectionResult",
    "TravelLoginType", "MAX_TOKEN_SIZE", "MAX_TOKEN_B64_SIZE",
    "hex_size_of_bigid",
    "TokenCreationError", "index_names", "protected_names", "spkcgraph",
    "dangerous_login_choices", "ActionUrl"
)

import enum
from collections import namedtuple

from rdflib.namespace import Namespace


spkcgraph = Namespace("https://spkcspider.net/static/schemes/spkcgraph#")
# Literal allows arbitary datatypes, use this and don't bind

hex_size_of_bigid = 16

MAX_TOKEN_SIZE = 90

if MAX_TOKEN_SIZE % 3 != 0:
    raise Exception("MAX_TOKEN_SIZE must be multiple of 3")

MAX_TOKEN_B64_SIZE = MAX_TOKEN_SIZE*4//3


class TokenCreationError(Exception):
    pass


ProtectionResult = namedtuple("ProtectionResult", ["result", "protection"])
ActionUrl = namedtuple("ActionUrl", ["url", "name"])
index_names = ["index", "fake_index"]
protected_names = ["index", "fake_index"]


class ProtectionType(str, enum.Enum):
    # receives: request, scope
    access_control = "a"
    # receives: request, scope, password
    authentication = "b"
    # c: former reliable, no meaning
    # protections which does not contribute to required_passes (404 only)
    no_count = "d"
    # protections which have side effects
    side_effects = "e"
    # forget about recovery, every recovery method is authentication
    # and will be misused this way
    # The only real recovery is by staff and only if there is a secret


class VariantType(str, enum.Enum):
    # a not assigned
    # use persistent token
    #  can be used for some kind of federation
    #  warning: if token is deleted persisted content is deleted
    persist = "b"
    # update content is without form/for form updates it is not rendered
    # required for still beeing able to update elemental parameters
    raw_update = "c"
    # raw_add not required, archieved by returning response

    # don't list as contentvariant for user (for computer only stuff)
    #  works like unlisted in info field, just for ContentVariants
    # don't appear in allowed_content in contrast to feature
    # => should either depend on feature or normal content
    unlisted = "d"
    # activates domain mode
    domain_mode = "e"
    # allow outside applications push content to spiders
    # adds by default unlisted attribute
    # appears in features of userComponent
    # counts as foreign content
    # don't list as content variant for user
    component_feature = "f"
    # same, but assigns to contents
    content_feature = "g"

    # is content unique for usercomponent
    # together with strength level 10: unique for user
    unique = "h"
    # can be used as anchor, hash is automatically embedded
    anchor = "i"


class TravelLoginType(str, enum.Enum):
    # no protection, login works as usual
    none = "a"
    # experimental, creates a fake subset view
    fake_login = "b"
    # wipe travel protected data and index as soon as login occurs
    # Note: noticable if shared contents are removed
    wipe = "c"
    # wipe user, self destruct user on login, Note: maybe noticable
    wipe_user = "d"


dangerous_login_choices = (
    TravelLoginType.fake_login.value,
    TravelLoginType.wipe.value,
    TravelLoginType.wipe_user.value
)



# Generated by Django 2.1.2 on 2018-10-07 18:31

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('spider_filets', '0005_auto_20180919_1329'),
    ]

    operations = [
        migrations.AddField(
            model_name='textfilet',
            name='preview_words',
            field=models.PositiveIntegerField(default=10, help_text='How many words should be used for preview?'),
        ),
    ]



# Generated by Django 2.1.5 on 2019-01-28 22:51

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0005_auto_20190128_2230'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='assignedcontent',
            name='nonce',
        ),
        migrations.RemoveField(
            model_name='usercomponent',
            name='nonce',
        ),
    ]



# Generated by Django 2.1 on 2018-08-30 13:52

import datetime
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import jsonfield.fields
import spkcspider.apps.spider.helpers
import spkcspider.apps.spider.models.content_base
import spkcspider.apps.spider.models.base
import spkcspider.apps.spider.models.protections


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='AssignedContent',
            fields=[
                ('id', models.BigAutoField(editable=False, primary_key=True, serialize=False)),
                ('nonce', models.SlugField(db_index=False, default=spkcspider.apps.spider.helpers.create_b64_token, max_length=120)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('deletion_requested', models.DateTimeField(blank=True, default=None, null=True)),
                ('info', models.TextField(editable=False, validators=[spkcspider.apps.spider.models.base.info_field_validator])),
                ('object_id', models.BigIntegerField(editable=False)),
                ('content_type', models.ForeignKey(editable=False, on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
            ],
        ),
        migrations.CreateModel(
            name='AssignedProtection',
            fields=[
                ('id', models.BigAutoField(primary_key=True, serialize=False)),
                ('data', jsonfield.fields.JSONField(default=dict)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('active', models.BooleanField(default=True)),
                ('instant_fail', models.BooleanField(default=False, help_text='Auth fails if test fails, stronger than required_passes\nWorks even if required_passes=0\nDoes not contribute to required_passes, ideal for side effects')),
            ],
        ),
        migrations.CreateModel(
            name='AuthToken',
            fields=[
                ('id', models.BigAutoField(editable=False, primary_key=True, serialize=False)),
                ('session_key', models.CharField(max_length=40, null=True)),
                ('token', models.SlugField(max_length=136)),
                ('created', models.DateTimeField(auto_now_add=True)),
            ],
        ),
        migrations.CreateModel(
            name='ContentVariant',
            fields=[
                ('id', models.BigAutoField(editable=False, primary_key=True, serialize=False)),
                ('ctype', models.CharField(max_length=10)),
                ('code', models.CharField(max_length=255)),
                ('name', models.SlugField(max_length=255, unique=True)),
            ],
        ),
        migrations.CreateModel(
            name='LinkContent',
            fields=[
                ('id', models.BigAutoField(editable=False, primary_key=True, serialize=False)),
                ('content', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='spider_base.AssignedContent')),
            ],
            options={
                'abstract': False,
                'default_permissions': [],
            },
        ),
        migrations.CreateModel(
            name='Protection',
            fields=[
                ('code', models.SlugField(db_index=False, max_length=10, primary_key=True, serialize=False)),
                ('ptype', models.CharField(default='b', max_length=10)),
            ],
        ),
        migrations.CreateModel(
            name='UserComponent',
            fields=[
                ('id', models.BigAutoField(editable=False, primary_key=True, serialize=False)),
                ('nonce', models.SlugField(db_index=False, default=spkcspider.apps.spider.helpers.create_b64_token, max_length=120)),
                ('public', models.BooleanField(default=False, help_text=(
                    "Is public? Is listed and searchable?<br/>"
                    "Note: This field is maybe not deactivatable"
                    "because of assigned content"
                ))),
                ('required_passes', models.PositiveIntegerField(default=0, help_text='How many protection must be passed? Set greater 0 to enable protection based access')),
                ('name', models.SlugField(db_index=False, help_text='\nName of the component.<br/>\nNote: there are special named components\nwith different protection types and scopes.<br/>\nMost prominent: "index" for authentication\n')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('token_duration', models.DurationField(default=datetime.timedelta(7))),
                ('deletion_requested', models.DateTimeField(default=None, null=True)),
                ('user', models.ForeignKey(editable=False, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name='UserInfo',
            fields=[
                ('id', models.BigAutoField(editable=False, primary_key=True, serialize=False)),
                ('used_space', models.BigIntegerField(default=0, editable=False)),
                ('allowed_content', models.ManyToManyField(editable=False, related_name='_userinfo_allowed_content_+', to='spider_base.ContentVariant')),
                ('user', models.OneToOneField(editable=False, on_delete=django.db.models.deletion.CASCADE, related_name='spider_info', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'default_permissions': [],
            },
        ),
        migrations.AddField(
            model_name='authtoken',
            name='usercomponent',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='authtokens', to='spider_base.UserComponent'),
        ),
        migrations.AddField(
            model_name='assignedprotection',
            name='protection',
            field=models.ForeignKey(editable=False, limit_choices_to=spkcspider.apps.spider.models.protections.get_limit_choices_assigned_protection, on_delete=django.db.models.deletion.CASCADE, related_name='assigned', to='spider_base.Protection'),
        ),
        migrations.AddField(
            model_name='assignedprotection',
            name='usercomponent',
            field=models.ForeignKey(editable=False, on_delete=django.db.models.deletion.CASCADE, related_name='protections', to='spider_base.UserComponent'),
        ),
        migrations.AddField(
            model_name='assignedcontent',
            name='ctype',
            field=models.ForeignKey(editable=False, null=True, on_delete=django.db.models.deletion.SET_NULL, to='spider_base.ContentVariant'),
        ),
        migrations.AddField(
            model_name='assignedcontent',
            name='usercomponent',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='contents', to='spider_base.UserComponent'),
        ),
        migrations.AlterUniqueTogether(
            name='usercomponent',
            unique_together={('user', 'name')},
        ),
        migrations.AlterUniqueTogether(
            name='authtoken',
            unique_together={('usercomponent', 'token')},
        ),
        migrations.AlterUniqueTogether(
            name='assignedprotection',
            unique_together={('protection', 'usercomponent')},
        ),
    ]

    operations.append(
        migrations.AlterUniqueTogether(
            name='assignedcontent',
            unique_together={('content_type', 'object_id')},
        )
    )



__all__ = [
    "clean_graph", "get_hashob", "validate_request_default",
    "verify_tag_default", "domain_auth"
]

import logging
from urllib.parse import parse_qs, urlencode

from django.conf import settings
from django.urls import reverse

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes

from rdflib import XSD, URIRef, Literal

import requests
import certifi

from spkcspider.apps.spider.constants import spkcgraph
from spkcspider.apps.spider.helpers import create_b64_token


def validate_request_default(request, form):
    if not form.is_valid():
        return False

    return True


def verify_tag_default(tag):
    tag.save()


def domain_auth(source, hostpart):
    GET = parse_qs(source.get_params)
    if "token" not in GET:
        GET["token"] = "prefer"
    GET["intention"] = "domain"
    GET["referrer"] = "{}{}".format(
        hostpart,
        reverse("spider_verifier:create")
    )
    source.update_secret = create_b64_token()
    source.save(update_fields=["update_secret"])
    GET["payload"] = urlencode(
        {
            "url": source.url,
            "update_secret": source.update_secret
        }
    )

    url = "{}?{}".format(
        source.url, urlencode(GET)
    )
    ret = True
    try:
        resp = requests.get(
            url, verify=certifi.where(),
            timeout=settings.VERIFIER_REQUESTS_TIMEOUT
        )
        resp.raise_for_status()
    except Exception:
        if settings.DEBUG:
            logging.exception("domain_auth failed")
        ret = False
    # other or own update_secret was successful or url has problems
    source.refresh_from_db()
    return ret


def clean_graph(mtype, graph, start, source, hostpart):
    if not mtype:
        return None
    elif "UserComponent" in mtype:
        return "list"
    elif "SpiderTag" in mtype:
        if (
            URIRef(start),
            spkcgraph["features"],
            Literal("verify", datatype=XSD.string)
        ) in graph:
            return "layout_cb"
        ret = "layout"
        if domain_auth(source, hostpart):
            graph.set(
                (
                    start,
                    spkcgraph["action:view"],
                    Literal(source.get_url(), datatype=XSD.anyURI)
                )
            )
            ret = "layout_cb"
        return ret
    else:
        return "content"


def get_hashob():
    return hashes.Hash(
        getattr(
           settings, "VERIFICATION_HASH_ALGORITHM",
           settings.SPIDER_HASH_ALGORITHM
        ),
        backend=default_backend()
    )



# Generated by Django 2.1.7 on 2019-03-17 14:05

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0008_auto_20190225_2034'),
    ]

    operations = [
        migrations.AddField(
            model_name='assignedcontent',
            name='allow_domain_mode',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='assignedcontent',
            name='attached_to_content',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='attached_contents', to='spider_base.AssignedContent'),
        ),
        migrations.AddField(
            model_name='assignedcontent',
            name='features',
            field=models.ManyToManyField(blank=True, limit_choices_to=models.Q(ctype__contains='g'), related_name='feature_for_contents', to='spider_base.ContentVariant'),
        ),
        migrations.AlterField(
            model_name='usercomponent',
            name='features',
            field=models.ManyToManyField(blank=True, limit_choices_to=models.Q(ctype__contains='f'), related_name='feature_for_components', to='spider_base.ContentVariant'),
        ),
    ]



__all__ = [
    "LinkForm", "TravelProtectionForm"
]

from django import forms
from django.conf import settings
from django.db import models
from django.contrib.auth.hashers import (
    make_password,
)
from django.utils.translation import gettext_lazy as _

from ..constants import dangerous_login_choices
from ..models import LinkContent, TravelProtection
from ..helpers import create_b64_token
from ..widgets import HTMLWidget


_extra = '' if settings.DEBUG else '.min'

PROTECTION_CHOICES = [
    ("none", _("No self-protection (not recommended)")),
    ("pw", _("Password")),
    ("token", _("Token"))
]

KEEP_CHOICES = [("keep", _("Keep protection"))] + PROTECTION_CHOICES


_self_protection = _("""Disallows user to disable travel protection if active.
 Can be used in connection with "secret" to allow unlocking via secret""")


class LinkForm(forms.ModelForm):

    class Meta:
        model = LinkContent
        fields = ['content', 'push']

    def __init__(self, uc, request, **kwargs):
        super().__init__(**kwargs)
        # if self.instance.associated:
        #     if "\x1eanchor\x1e" in self.instance.associated:
        #         self.fields["content"].disabled = True
        q = self.fields["content"].queryset
        travel = TravelProtection.objects.get_active()
        self.fields["content"].queryset = q.filter(
            strength__lte=uc.strength
        ).exclude(usercomponent__travel_protected__in=travel)
        # component auth should limit links to visible content
        # read access outside from component elsewise possible
        if request.user != uc.user and not request.is_staff:
            q = self.fields["content"].queryset
            self.fields["content"].queryset = q.filter(
                models.Q(usercomponent=uc) |
                models.Q(referenced_by__usercomponent=uc)
            )


class TravelProtectionForm(forms.ModelForm):
    uc = None
    is_fake = None
    _password = None
    password = forms.CharField(
        label=_("Old Password"),
        strip=False,
        widget=forms.PasswordInput,
    )
    self_protection = forms.ChoiceField(
        label=_("Self-protection"), help_text=_(_self_protection),
        initial="None", choices=PROTECTION_CHOICES
    )
    token_arg = forms.CharField(
        label=_("Token"),
        help_text=_("please export/copy for disabling travel mode"),
        widget=HTMLWidget
    )
    new_pw = forms.CharField(
        initial="", required=False, label=_("New Password"),
        widget=forms.PasswordInput(),
    )
    new_pw2 = forms.CharField(
        initial="", required=False, label=_("New Password (Retype)"),
        widget=forms.PasswordInput(),
    )

    class Meta:
        model = TravelProtection
        fields = [
            "active", "start", "stop", "login_protection", "disallow"
        ]

    class Media:
        js = [
            'spider_base/travelprotection.js'
        ]

    def __init__(self, request, uc, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.uc = uc
        self.request = request

        q = models.Q(
            user=self.uc.user,
            strength__lt=10  # don't disclose other index
        ) & ~models.Q(
            travel_protected__in=TravelProtection.objects.get_active()
        ) & ~models.Q(
            public=True  # this would easily expose the travel mode
        )

        if not getattr(settings, "DANGEROUS_TRAVEL_PROTECTIONS", False):
            self.fields["login_protection"].choices = \
                filter(
                    lambda x: x[0] not in dangerous_login_choices,
                    self.fields["login_protection"].choices
                )

        self.fields["disallow"].queryset = \
            self.fields["disallow"].queryset.filter(q)

        # elif self.travel_protection.is_active:
        #    for f in self.fields:
        #        f.disabled = True
        self.fields["token_arg"].initial = create_b64_token(30)

        if not self.instance.active or not self.instance.hashed_secret:
            del self.fields["password"]

        if self.instance.hashed_secret:
            self.fields["self_protection"].choices = KEEP_CHOICES

        # doesn't matter if it is same user, lazy
        travel = TravelProtection.objects.get_active(no_stop=True)
        self.fields["disallow"].queryset = self.fields["disallow"].queryset.\
            filter(
                ~models.Q(travel_protected__in=travel),
                user=self.uc.user, strength__lt=10
            )

    def is_valid(self):
        isvalid = super().is_valid()
        if not getattr(self, "cleaned_data", None):
            return False
        if self.instance.active and "password" in self.fields:
            isvalid = self.instance.check_password(
                self.cleaned_data["password"]
            )
        return isvalid

    def clean(self):
        ret = super().clean()
        if ret["self_protection"] == "none":
            self._password = False
        if ret["self_protection"] == "pw":
            if self.cleaned_data["new_pw"] != self.cleaned_data["new_pw2"]:
                self.add_error("new_pw", forms.ValidationError(
                    _("The two password fields didn't match.")
                ))
                self.add_error("new_pw2", forms.ValidationError(
                    _("The two password fields didn't match.")
                ))
            else:
                self._password = self.cleaned_data["new_pw"]
        elif ret["self_protection"] == "token":
            self._password = self.fields["token_arg"].initial
        return ret

    def save(self, commit=True):
        self.instance.is_fake = self.request.session.get("is_fake", False)
        if self._password:
            self.instance.hashed_secret = make_password(self._password)
        if self._password is False:
            self.instance.hashed_secret = None
        return super().save(commit=commit)



from django import template
from django.urls import reverse
from django.utils import timezone
from django.urls.exceptions import NoReverseMatch

register = template.Library()


@register.filter()
def is_not_or_space(value):
    if not value:
        return True
    return isinstance(value, str) and value.isspace()


@register.simple_tag(takes_context=True)
def current_url(context):
    try:
        return reverse(
            "{}:{}".format(
                context["request"].resolver_match.namespace,
                context["request"].resolver_match.url_name,
            ),
            args=context["request"].resolver_match.args,
            kwargs=context["request"].resolver_match.kwargs
        )
    except NoReverseMatch:
        return reverse(
            context["request"].resolver_match.url_name,
            args=context["request"].resolver_match.args,
            kwargs=context["request"].resolver_match.kwargs
        )


@register.simple_tag(takes_context=True)
def list_own_content(context):
    ucname = "index"
    if context["request"].session.get("is_fake", False):
        ucname = "fake_index"
    if context["request"].user.is_authenticated:
        return context["request"].user.usercomponent_set.get(
            name=ucname
        ).get_absolute_url()
    return ""


@register.simple_tag(takes_context=True)
def update_component(context, name):
    ucname = name
    if ucname == "index" and context["request"].session.get("is_fake", False):
        ucname = "fake_index"
    if context["request"].user.is_authenticated:
        uc = context["request"].user.usercomponent_set.only("token").get(
            name=ucname
        )
        return reverse("spider_base:ucomponent-update", kwargs={
            "token": uc.token
        })
    return ""


@register.simple_tag(takes_context=True)
def reverse_get(context, name, **kwargs):
    """ Works only if hostpart and spider_GET is available """
    return "{}{}?{}".format(
        context["hostpart"],
        reverse(name, kwargs=kwargs),
        context["spider_GET"].urlencode()
    )


@register.simple_tag()
def concat_string(*args):
    return "".join(args)


@register.simple_tag()
def expires_delta(expires):
    return expires-timezone.now()


@register.simple_tag()
def token_expires(usercomponent, token):
    return token.created + usercomponent.token_duration



# Generated by Django 2.1.4 on 2019-01-03 01:56

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0003_remove_usercomponent_avail_features'),
    ]

    operations = [
        migrations.AlterField(
            model_name='assignedcontent',
            name='ctype',
            field=models.ForeignKey(editable=False, null=True, on_delete=django.db.models.deletion.SET_NULL, to='spider_base.ContentVariant'),
        ),
    ]



# Generated by Django 2.1.5 on 2019-01-17 12:56

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0010_usercomponent_can_auth'),
    ]

    operations = [
        migrations.AddField(
            model_name='usercomponent',
            name='primary_anchor',
            field=models.ForeignKey(blank=True, limit_choices_to=models.Q(info__contains='\nanchor\n'), null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='spider_base.AssignedContent'),
        ),
        migrations.AlterField(
            model_name='authtoken',
            name='persist',
            field=models.BigIntegerField(blank=True, db_index=True, default=-1),
        ),
    ]



# Generated by Django 2.1.5 on 2019-01-09 19:12

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0005_auto_20190109_1505'),
    ]

    operations = [
        migrations.AddField(
            model_name='assignedcontent',
            name='priority',
            field=models.SmallIntegerField(blank=True, default=0),
        ),
    ]



# Generated by Django 2.1.5 on 2019-01-09 19:27

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0006_assignedcontent_priority'),
    ]

    operations = [
        migrations.AddField(
            model_name='linkcontent',
            name='push',
            field=models.BooleanField(blank=True, default=False, help_text='Push Link to top.'),
        ),
    ]



from django import template

register = template.Library()


@register.simple_tag(takes_context=True)
def list_parameters(context, name, *args):
    ret = set(context["request"].GET.getlist(name))
    ret.update(context["request"].POST.getlist(name))
    for i in args:
        ret.update(i)
    return ret


@register.simple_tag(takes_context=True)
def get_searchpath(context, page=None, pagename="page"):
    if "searchpath" in context:
        path = context["searchpath"]
    else:
        path = context["request"].path
    san_get = ""
    if "spider_GET" in context:
        san_get = context["spider_GET"].urlencode()
    if page:
        if san_get:
            "{}&{}={}".format(
                san_get,
                pagename,
                page
            )
        else:
            san_get = "{}={}".format(pagename, page)
    if path[-1] == "?":
        return "{}{}".format(
            path,
            san_get
        )
    return "{}?{}".format(
        path,
        san_get
    )



# Generated by Django 2.1.1 on 2018-09-19 13:29

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('spider_filets', '0004_auto_20180821_0021'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='filefilet',
            name='add',
        ),
        migrations.RemoveField(
            model_name='filefilet',
            name='modified',
        ),
    ]



# Generated by Django 2.1.2 on 2018-11-02 18:26

from django.db import migrations
import jsonfield.fields


class Migration(migrations.Migration):

    dependencies = [
        ('spider_verifier', '0002_auto_20181031_2121'),
    ]

    operations = [
        migrations.AddField(
            model_name='dataverificationtag',
            name='linked_hashes',
            field=jsonfield.fields.JSONField(blank=True, default=dict),
        ),
    ]
